;
; static char sccs_id[] = "@(#)fchk32.asm	1.3  1/13/95  10:07:22";
;
; This code is being published by Intel to users of the Pentium(tm)
; processor.  Recipients are authorized to copy, modify, compile, use and
; distribute the code.
;
; Intel makes no warranty of any kind with regard to this code, including
; but not limited to, implied warranties or merchantability and fitness for
; a particular purpose. Intel assumes no responsibility for any errors that
; may appear in this code.
;
; No patent licenses are granted, express or implied.
;
;
;  PRELIMINARY VERSION of the software patch for the floating
;  point divide.
;
;

.386
.387

DATA32  SEGMENT DWORD USE32 PUBLIC 'DATA'

fdiv_risc_table	DB	0, 1, 0, 0, 4, 0, 0, 7, 0, 0, 10, 0, 0, 13, 0, 0
fdiv_scale_1	DD	03f700000h		;0.9375
fdiv_scale_2	DD	03f880000h		;1.0625
one_shl_63	DD	05f000000h

;
;  This is the flag that is tested by all the compiler branches
;  before the divide instruction.
;
;	0 = unknown whether running on a part with the flaw
;	1 = running on a part with no flaw
;	-1= running on a part with the flaw
;

	public	_fdiv_chk_flag
_fdiv_chk_flag 	DD	0h

;
;  The following are initializers for known values that cause
;  the divide problem.
;

test_dividend	DB	0,0,0,0c0h,07eh,01h,050h,041h	 ; x = 4195835.0
test_divisor	DB	0,0,0,080h,0ffh,0ffh,047h,041h   ; y = 3145727.0
one		DB	0,0,0,0,0,0,0f0h,03fh

DATA32  ENDS

;
;  Stack variables for divide routines.
;

DENOM		EQU	0
NUMER		EQU	12
PATCH_CW 	EQU	32
PREV_CW		EQU	28
STACK_SIZE	EQU	32

MAIN_DENOM	EQU	4
MAIN_NUMER	EQU	16


;
;  Stack frame locations for  C/C++.
;  Single real is pushed as 4 bytes, Double real is pushed as 8 bytes and
;  Long Double (extended) real is pushed as 16 bytes. Needs to be changed the
;  following definitions for different calling conventions.
;

;
; Definition for Single Real
;
SINGLE1		EQU	4 + STACK_SIZE
SINGLE2		EQU	8  + STACK_SIZE

;
; Definition for Double Real
;
DOUBLE1		EQU	4  + STACK_SIZE
DOUBLE2		EQU	12  + STACK_SIZE

;
; Definition for Long Double Real (Extended)
;
LDOUBLE1	EQU	4 + STACK_SIZE
LDOUBLE2	EQU	20  + STACK_SIZE

SRESULT		EQU	12  + STACK_SIZE
DRESULT		EQU	20 + STACK_SIZE

ONESMASK	EQU	0e000000h

CODE32  SEGMENT DWORD USE32 PUBLIC 'CODE'
CODE32  ENDS

DATA32  SEGMENT DWORD USE32 PUBLIC 'DATA'
DATA32  ENDS

CONST32 SEGMENT DWORD USE32 PUBLIC 'CONST'
CONST32 ENDS

BSS32   SEGMENT DWORD USE32 PUBLIC 'BSS'
BSS32   ENDS

DGROUP  GROUP CONST32, BSS32, DATA32



CODE32  SEGMENT   DWORD USE32 PUBLIC 'CODE'


	assume cs:CODE32, ds:FLAT, es:FLAT, ss:nothing



					; In this implementation the
					; fdiv_main_routine is called,
					; therefore all the stack frame
					; locations are adjusted for the
					; return pointer.

fdiv_main_routine PROC	NEAR

	fld     tbyte ptr [esp+MAIN_NUMER]   ; load the numerator
	fld	tbyte ptr [esp+MAIN_DENOM]	; load the denominator
retry:

;  The following three lines test for denormals and zeros.
;  A denormal or zero has a 0 in the explicit digit to the left of the
;  binary point.  Since that bit is the high bit of the word, adding
;  it to itself will produce a carry if and only if the number is not
;  denormal or zero.
;
	mov	eax, [esp+MAIN_DENOM+4]	; get mantissa bits 32-64
	add	eax,eax			; shift the one's bit onto carry
	jnc	denormal		; if no carry, we're denormal

;  The following three lines test the three bits after the four bit 
;  pattern (1,4,7,a,d).  If these three bits are not all one, then
;  the denominator cannot expose the flaw.  This condition is tested by
;  inverting the bits and testing that all are equal to zero afterward.

	xor	eax, ONESMASK		; invert the bits that must be ones
	test	eax, ONESMASK		; and make sure they are all ones
	jz	scale_if_needed		; if all are one scale numbers
	fdivp	st(1), st		; use of hardware is OK.
	ret

;
;  Now we test the four bits for one of the five patterns.
;
scale_if_needed:
	shr	eax, 28			; keep first 4 bits after point
	cmp	byte ptr fdiv_risc_table[eax], 0	; check for (1,4,7,a,d)
	jnz	divide_scaled		; are in potential problem area
	fdivp	st(1), st		; use of hardware is OK.
	ret

divide_scaled:
	mov	eax, [esp + MAIN_DENOM+8]	; test denominator exponent
	and	eax, 07fffh             ; if pseudodenormal ensure that only
	jz	invalid_denom		; invalid exception flag is set
	cmp	eax, 07fffh             ; if NaN or infinity  ensure that only
	je	invalid_denom		; invalid exception flag is set
;
;  The following six lines turn off exceptions and set the
;  precision control to 80 bits.  The former is necessary to
;  force any traps to be taken at the divide instead of the scaling
;  code.  The latter is necessary in order to get full precision for
;  codes with incoming 32 and 64 bit precision settings.  If
;  it can be guaranteed that before reaching this point, the underflow
;  exception is masked and the precision control is at 80 bits, these
;  six lines can be omitted.
;
	fnstcw	[esp+PREV_CW]		; save caller's control word
	mov	eax, [esp+PREV_CW] 
	or	eax, 033fh		; mask exceptions, pc=80
	and	eax, 0f3ffh		; set rounding mode to nearest
	mov	[esp+PATCH_CW], eax
	fldcw	[esp+PATCH_CW]		; mask exceptions & pc=80

;  The following lines check the numerator exponent before scaling.
;  This in order to prevent undeflow when scaling the numerator,
;  which will cause a denormal exception flag to be set when the
;  actual divide is preformed. This flag would not have been set
;  normally. If there is a risk of underflow, the scale factor is
;  17/16 instead of 15/16.
;
 	mov	eax, [esp+MAIN_NUMER+8]	; test numerator exponent
 	and	eax, 07fffh
 	cmp	eax, 00001h
 	je	small_numer
 
	fmul	fdiv_scale_1		; scale denominator by 15/16
	fxch
	fmul	fdiv_scale_1		; scale numerator by 15/16
	fxch

;
;  The next line restores the users control word.  If the incoming
;  control word had the underflow exception masked and precision
;  control set to 80 bits, this line can be omitted.
;

	fldcw	[esp+PREV_CW]		; restore caller's control word
	fdivp	st(1), st		; use of hardware is OK.
	ret

small_numer:
	fmul	fdiv_scale_2		; scale denominator by 17/16
	fxch
	fmul	fdiv_scale_2		; scale numerator by 17/16
	fxch

;
;  The next line restores the users control word.  If the incoming
;  control word had the underflow exception masked and precision
;  control set to 80 bits, this line can be omitted.
;

	fldcw	[esp+PREV_CW]		; restore caller's control word
	fdivp	st(1), st		; use of hardware is OK.
	ret

denormal:
	mov	eax, [esp+MAIN_DENOM]	; test for whole mantissa == 0
	or	eax, [esp+MAIN_DENOM+4]	; test for whole mantissa == 0
	jnz	denormal_divide_scaled	; denominator is not zero
invalid_denom:				; zero or invalid denominator
	fdivp	st(1), st		; use of hardware is OK.
	ret

denormal_divide_scaled:
	mov	eax, [esp + MAIN_DENOM + 8]	; get exponent 
	and	eax, 07fffh		; check for zero exponent
	jnz	invalid_denom		; 
;
;  The following six lines turn off exceptions and set the
;  precision control to 80 bits.  The former is necessary to
;  force any traps to be taken at the divide instead of the scaling
;  code.  The latter is necessary in order to get full precision for
;  codes with incoming 32 and 64 bit precision settings.  If
;  it can be guaranteed that before reaching this point, the underflow
;  exception is masked and the precision control is at 80 bits, these
;  six lines can be omitted.
;

	fnstcw	[esp+PREV_CW]		; save caller's control word
	mov	eax, [esp+PREV_CW] 
	or	eax, 033fh		; mask exceptions, pc=80
	and	eax, 0f3ffh		; set rounding mode to nearest
	mov	[esp+PATCH_CW], eax
	fldcw	[esp+PATCH_CW]		; mask exceptions & pc=80

	mov	eax, [esp + MAIN_NUMER +8]	; test numerator exponent
	and	eax, 07fffh		; check for denormal numerator
	je	denormal_numer	
	cmp	eax, 07fffh		; NaN or infinity
	je	invalid_numer
	mov	eax, [esp + MAIN_NUMER + 4]	; get bits 32..63 of mantissa
	add	eax, eax		; shift the first bit into carry
	jnc	invalid_numer		; if there is no carry, we have an
					; invalid numer
	jmp	numer_ok

denormal_numer:
	mov	eax, [esp + MAIN_NUMER + 4]	; get bits 32..63 of mantissa
	add	eax, eax		; shift the first bit into carry
	jc	invalid_numer		; if there is a carry, we have an
					; invalid numer
	
numer_ok:
	fxch
	fstp	st			; pop numerator
	fld	st			; make copy of denominator
	fmul	dword ptr[one_shl_63]	; make denominator not denormal
	fstp	tbyte ptr [esp+MAIN_DENOM]
	fld	tbyte ptr [esp+MAIN_NUMER]	; load numerator
	fxch				; restore proper order
	fwait

;  The next line restores the users control word.  If the incoming
;  control word had the underflow exception masked and precision
;  control set to 80 bits, this line can be omitted.
;

	fldcw	[esp+PREV_CW]		; restore caller's control word
	jmp	retry			; start the whole thing over

invalid_numer:
;
;  The next line restores the users control word.  If the incoming
;  control word had the underflow exception masked and precision
;  control set to 80 bits, this line can be omitted.
;
	fldcw	[esp + PREV_CW]
	fdivp	st(1), st		; use of hardware is OK.
	ret

fdiv_main_routine	ENDP

;
; int fdiv_detect(void)
;

	public _FDIV_DETECT
_FDIV_DETECT PROC NEAR
	fld	qword ptr [test_dividend]
	fwait
	fdiv	qword ptr [test_divisor] ; no_error = ((x - (x / y)*y) < 1.0
	fwait
	fmul	qword ptr [test_divisor]
	fwait
	fsubr	qword ptr [test_dividend]
	fwait
	fcomp	qword ptr [one]
	fwait
	fstsw	ax
	fwait
	and	eax, 0100h              ; 0: uninitialized
	shr	eax, 7
	sub	eax, 1
	mov	_fdiv_chk_flag, eax
	ret
_FDIV_DETECT ENDP

;
;  The calling convention is DIV_CHK(double,double,double*)
;
;  This routine assumes a calling convention of double real being pushed as
;  8 bytes.
;
	public  _DIV_CHK
_DIV_CHK     PROC    NEAR
	sub     esp, STACK_SIZE
	fld	qword ptr [esp+DOUBLE1]
	fstp	tbyte ptr [esp+NUMER]
	fld	qword ptr [esp+DOUBLE2]
	fstp	tbyte ptr [esp+DENOM]
	call	fdiv_main_routine
	mov	eax, [esp+DRESULT]
	fstp	qword ptr [eax]
	add	esp, STACK_SIZE
	ret
_DIV_CHK     ENDP

;
;  The calling convention is FDIV_CHK(float,float,float*)
;
;  This routine assumes a calling convention of single real being pushed as
;  4 bytes.
;
	public  _FDIV_CHK
_FDIV_CHK     PROC    NEAR
	sub     esp, STACK_SIZE
	fld	dword ptr [esp+SINGLE1]
	fstp	tbyte ptr [esp+NUMER]
	fld	dword ptr [esp+SINGLE2]
	fstp	tbyte ptr [esp+DENOM]
	call	fdiv_main_routine 
	mov	eax, [esp+SRESULT]
	fstp	dword ptr [eax]
	add	esp, STACK_SIZE
	ret
_FDIV_CHK     ENDP

;
;  The calling convention is LDIV_CHK(long double, long double) and
;  returns a long double, ie, result is in th top of FP Stack (ST).
;
;  This routine assumes a calling convention of long double real being 
;  pushed as  16 bytes.
;
	public  _LDIV_CHK
_LDIV_CHK     PROC    NEAR
	sub     esp, STACK_SIZE
	fld	tbyte ptr [esp+LDOUBLE1]
	fstp	tbyte ptr [esp+NUMER]
	fld	tbyte ptr [esp+LDOUBLE2]
	fstp	tbyte ptr [esp+DENOM]
	call	fdiv_main_routine 
	add	esp, STACK_SIZE
	ret
_LDIV_CHK     ENDP

CODE32  ENDS

	end
